/*
 * Copyright (C) 2011 Kenichi Takahashi
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

/* OpenSL ES を使用する際に必要なヘッダ */
#include <SLES/OpenSLES.h>
/* OpenSL ES の Android 拡張機能を使用する際に必要なヘッダ */
#include <SLES/OpenSLES_Android.h>

/*
 * native_activity.h をダイレクトに使っても良いのですが、android_native_app_glue.h
 * という helper ライブラリが用意されていますので、これを使うことにします。
 * それにより Activity のライフサイクルの各イベントの処理の実装が簡単になります。
 */
#include <android_native_app_glue.h>
#include <android/asset_manager.h>

#include <assert.h>
#include <errno.h>
#include <sys/time.h>
#include <time.h>
#include <android/log.h>

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#define  LOG_TAG    "SD7sample"

struct engine {
    struct android_app* app;
    int needToDraw; /* 描画の必要を示すフラグ */
    int playing; /* 音声再生状態 */
    /* タップ座標の保持 */
    float xpos; /* X座標 */
    float ypos; /* Y座標 */
};

/* engine interfaces */
static SLObjectItf engineObject = NULL;
static SLEngineItf engineEngine;

/* output mix interfaces */
static SLObjectItf outputMixObject = NULL;

/* file descriptor player interfaces */
static SLObjectItf fdPlayerObject = NULL;
static SLPlayItf fdPlayerPlay;
static SLSeekItf fdPlayerSeek;

static void setPlayingAssetAudioPlayer(struct engine *engine, int isPlaying);
static void playerCallback(SLPlayItf caller, void *context, SLuint32 event);

/*
 * 画面のバッファを指定色でクリアします。
 *
 * 引数：
 * buffer ....... 画面のバッファ
 * cloerColor ... クリアする色 (RGB565)
 *
 */
static void clearFrame (ANativeWindow_Buffer* buffer, uint16_t clearColor)
{
    /* 受け取ったバッファのピクセル毎にアクセスするためのポインタを取り出します。*/
    uint16_t* pixels = (uint16_t *)(buffer->bits);

    int x;
    int y;
    /* 画面縦方向のループの定義です。buffer->height が画面縦方向のサイズを示しています。 */
    for (y = 0; y < buffer->height; y++) {
      /* 画面横方向のループ定義です。buffer->width が画面横方向のサイズを示しています。*/
      for (x = 0; x < buffer->width; x++) {
	  /* クリア色で埋めます。 */
	  *pixels++ = clearColor;
      }
    }
}

/*
 * 矩形を描画します。　
 *
 * 引数：
 * buffer ... 画面のバッファ
 * color .... 描画色 (RGB565)
 * cx ....... 矩形の左上の X 座標（単位：ピクセル）
 * cy ....... 矩形の左上の Y 座標（単位：ピクセル）
 * size ..... 矩形の縦横のサイズ（単位：ピクセル）
 *
 */
static void drawRect (ANativeWindow_Buffer* buffer, uint16_t color, int cx, int cy, int size)
{
    /* 受け取ったバッファのピクセル毎にアクセスするためのポインタを取り出します。*/
    uint16_t* pixels = (uint16_t *)(buffer->bits);

    int x;
    int y;
    /* 縦方向のサイズ分のループです。 */
    for (y = 0; y < size; y++) {
      /* 横方向の描画の開始位置を求めます。 */
      uint16_t *pix = pixels + buffer->width * (cy + y) + cx;
      /* 横方向のサイズ分のループです。 */
      for (x = 0; x < size; x++) {
	  /* 描画色で埋めます。 */
	  *pix++ = color;
      }
    }
}

/*
 * 画面の描画を行う必要がある時に呼び出される関数
 */
static void engine_draw_frame(struct engine* engine)
{
    if (engine->app->window == NULL) {
        // No window.
        return;
    }

    /*
     * 画面バッファ情報
     * ここに画面描画に必要な幅,高さ,バッファへのポインタ等の情報が入ります。
     * usr/include/android/native_window.h に定義されています。
     */
    ANativeWindow_Buffer buffer;

    /* 画面をロックして buffer に画面描画のためのバッファを関連づけます。 */
    if (ANativeWindow_lock(engine->app->window, &buffer, NULL) < 0) {
        __android_log_print (ANDROID_LOG_WARN, LOG_TAG, "Unable to lock window buffer");
        return;
    }

    /* 音声の再生状態で処理を切り替える */
    if (engine->playing) {
      /* 再生ならタッチ位置に矩形を描画する */
      drawRect(&buffer, 0xffff, (int)engine->xpos, (int)engine->ypos, 50);
    } else {
      /* 停止なら画面をクリアする */
      clearFrame(&buffer, 0);
    }

    /* 画面のロックを解除し、画面に描画結果を反映させます */
    ANativeWindow_unlockAndPost(engine->app->window);
}

/*
 * アプリ終了時に呼び出される関数の定義です。
 */
static int engine_term_display(struct engine* engine) {
    engine->needToDraw = 0;
    engine->playing = 0;

    /* OpenSL ES 関連のオブジェクトを解放します。 */
    /* プレイヤーオブジェクト */
    if (fdPlayerObject != NULL) {
        (*fdPlayerObject)->Destroy(fdPlayerObject);
        fdPlayerObject = NULL;
        fdPlayerPlay = NULL;
        fdPlayerSeek = NULL;
    }
    /* 出力先オブジェクト */
    if (outputMixObject != NULL) {
        (*outputMixObject)->Destroy(outputMixObject);
        outputMixObject = NULL;
    }
    /* エンジンオブジェクト */
    if (engineObject != NULL) {
        (*engineObject)->Destroy(engineObject);
        engineObject = NULL;
        engineEngine = NULL;
    }
}

/*
 * 入力イベントがあった時に呼び出される関数
 */
static int32_t engine_handle_input(struct android_app* app, AInputEvent* event)
{
    struct engine* engine = (struct engine*)app->userData;
    if (AInputEvent_getType(event) == AINPUT_EVENT_TYPE_MOTION) {
        /* 画面のタップ操作の場合 */
        __android_log_print (ANDROID_LOG_INFO, LOG_TAG,
		"Motion event: action=%d flag=%d metaState=0x%x",
                AMotionEvent_getAction(event),
                AMotionEvent_getFlags(event),
	        AMotionEvent_getMetaState(event));
	if (AMotionEvent_getAction(event) == AMOTION_EVENT_ACTION_UP) {
	  /* タップ操作が画面に触れた後に手を離した状態の場合 */
	  /* タップ位置の取得 */
	  engine->xpos = AMotionEvent_getX(event, 0); /* X座標 */
	  engine->ypos = AMotionEvent_getY(event, 0); /* Y座標 */
	  __android_log_print (ANDROID_LOG_INFO, LOG_TAG,
			       "pos = (%f, %f)", 
			       engine->xpos, engine->ypos);
	  /* プレイヤーを再生状態にします */
	  setPlayingAssetAudioPlayer(engine, 1);
	  /* 描画必要フラグをON */
	  engine->needToDraw = 1;
	  /* 再生中フラグをON */
	  engine->playing = 1;
	}
        return 1;
    } else if (AInputEvent_getType(event) == AINPUT_EVENT_TYPE_KEY) {
        /* キー入力操作の場合 */
        __android_log_print (ANDROID_LOG_INFO, LOG_TAG,
		"Key event: action=%d keyCode=%d metaState=0x%x",
                AKeyEvent_getAction(event),
                AKeyEvent_getKeyCode(event),
                AKeyEvent_getMetaState(event));
    }

    return 0;
}

/*
 * ウィンドウの初期化、終了等のイベントがあった時に呼び出される関数
 */
static void engine_handle_cmd(struct android_app* app, int32_t cmd)
{
    struct engine* engine = (struct engine*)app->userData;
    switch (cmd) {
        case APP_CMD_INIT_WINDOW:
            if (engine->app->window != NULL) {
                engine_draw_frame(engine);
            }
            break;
        case APP_CMD_TERM_WINDOW:
            engine_term_display(engine);
            break;
        case APP_CMD_LOST_FOCUS:
            engine->needToDraw = 0;
            engine_draw_frame(engine);
            break;
    }
}

/*
 * 音声再生用エンジンの生成
 */
void createSoundEngine()
{
    SLresult result;
    /* エンジンオブジェクトの生成 */
    result = slCreateEngine(&engineObject, 0, NULL, 0, NULL, NULL);
    assert(SL_RESULT_SUCCESS == result);

    /* 生成オブジェクトはリアライズすることで内部リソースが確保されて使用可能になります */
    result = (*engineObject)->Realize(engineObject, SL_BOOLEAN_FALSE);
    assert(SL_RESULT_SUCCESS == result);

    /*
     * エンジンインターフェースの取得
     * 関連する他のオブジェクトを生成するために必要です。
     */
    result = (*engineObject)->GetInterface(engineObject, SL_IID_ENGINE, &engineEngine);
    assert(SL_RESULT_SUCCESS == result);

    /* 出力先オブジェクトの生成 */
    result = (*engineEngine)->CreateOutputMix(engineEngine, &outputMixObject, 0, NULL, NULL);
    assert(SL_RESULT_SUCCESS == result);

    /* 生成オブジェクトはリアライズすることで内部リソースが確保されて使用可能になります */
    result = (*outputMixObject)->Realize(outputMixObject, SL_BOOLEAN_FALSE);
    assert(SL_RESULT_SUCCESS == result);
}

/*
 * プレイヤーからのコールバック関数
 * （再生が終了した時に呼び出されます）
 */
void playerCallback(SLPlayItf caller, void *context, SLuint32 event)
{
    struct engine *engine = (struct engine *)context;
    __android_log_print (ANDROID_LOG_INFO, LOG_TAG, "The end of the sound. [%d]", engine->playing);
    /* 再生状態フラグを停止にします */
    engine->playing = 0;
    /* 画面をクリアするために描画必要フラグをONにします */
    engine->needToDraw = 1;
}

// create asset audio player
int createAssetAudioPlayer(struct engine *engine)
{
    SLresult result;

    /* AssetManager のインスタンスの取得 */
    AAssetManager *amgr = engine->app->activity->assetManager;
    assert(NULL != amgr);
    /* ファイル名を指定して音声ファイルのassetをオープンします */
    AAsset* asset = AAssetManager_open(amgr, "foo.wav", AASSET_MODE_UNKNOWN);
    // the asset might not be found
    if (NULL == asset) {
        return 0;
    }

    /* オープンしたassetのファイルディスクリプタを取得します */
    off_t start, length;
    int fd = AAsset_openFileDescriptor(asset, &start, &length);
    assert(0 <= fd);
    /* オープンした asset をクローズします */
    AAsset_close(asset);

    /* 音声データの場所を設定 */
    SLDataLocator_AndroidFD loc_fd = {SL_DATALOCATOR_ANDROIDFD, fd, start, length};
    /* 音声データのフォーマット関連の設定 */
    SLDataFormat_MIME format_mime = {SL_DATAFORMAT_MIME, NULL, SL_CONTAINERTYPE_UNSPECIFIED};
    SLDataSource audioSrc = {&loc_fd, &format_mime};

    /* 出力先の設定 */
    SLDataLocator_OutputMix loc_outmix = {SL_DATALOCATOR_OUTPUTMIX, outputMixObject};
    SLDataSink audioSnk = {&loc_outmix, NULL};

    const SLInterfaceID ids[1] = {SL_IID_SEEK};
    const SLboolean req[1] = {SL_BOOLEAN_TRUE};
    /* Audio Player オブジェクトの生成 */
    result = (*engineEngine)->CreateAudioPlayer(engineEngine, &fdPlayerObject, &audioSrc, &audioSnk,
            1, ids, req);
    assert(SL_RESULT_SUCCESS == result);

    /* 生成オブジェクトはリアライズすることで内部リソースが確保されて使用可能になります */
    result = (*fdPlayerObject)->Realize(fdPlayerObject, SL_BOOLEAN_FALSE);
    assert(SL_RESULT_SUCCESS == result);

    /* プレイヤー インターフェースの取得 */
    result = (*fdPlayerObject)->GetInterface(fdPlayerObject, SL_IID_PLAY, &fdPlayerPlay);
    assert(SL_RESULT_SUCCESS == result);

    /* シーク インターフェースの取得 */
    result = (*fdPlayerObject)->GetInterface(fdPlayerObject, SL_IID_SEEK, &fdPlayerSeek);
    assert(SL_RESULT_SUCCESS == result);

    /* プレイヤーインターフェースにコールバック関数を登録します */
    result = (*fdPlayerPlay)->RegisterCallback(fdPlayerPlay, playerCallback, engine);
    assert(SL_RESULT_SUCCESS == result);

    /* コールバックは再生中にファイルの末尾に到達した際に呼ばれるよう設定します */
    result = (*fdPlayerPlay)->SetCallbackEventsMask(fdPlayerPlay,
						    SL_PLAYEVENT_HEADATEND
						    );
    assert(SL_RESULT_SUCCESS == result);

    return 1;
}

/*
 * 音声の再生／停止状態を切り替えます。
 * 引数：
 * engine ...... アプリケーション構造体
 * isPlaying ... 再生状態（0: 停止 1: 再生）
 */
void setPlayingAssetAudioPlayer(struct engine *engine, int isPlaying)
{
    SLresult result;

    // make sure the asset audio player was created
    if (NULL != fdPlayerPlay) {
        /* 再生位置を先頭に移動します */
        result = (*(fdPlayerSeek))->SetPosition(fdPlayerSeek, 0, SL_SEEKMODE_FAST);
        assert(SL_RESULT_SUCCESS == result);
        /* 再生状態を設定します */
        result = (*fdPlayerPlay)->SetPlayState(fdPlayerPlay, isPlaying ?
            SL_PLAYSTATE_PLAYING : SL_PLAYSTATE_PAUSED);
        assert(SL_RESULT_SUCCESS == result);
    }
}

/*
 * android_main(struct android_app* state) がエントリの関数となり、
 * アプリが実行された際はここから処理が始まります。
 */
void android_main(struct android_app* state)
{
    struct engine engine;

    // Make sure glue isn't stripped.
    app_dummy();

    memset(&engine, 0, sizeof(engine));
    state->userData = &engine;
    /* 引数で受け取った state に対してイベントが発生した際に呼ばれる関数を結びつけています。 */
    state->onAppCmd = engine_handle_cmd;
    state->onInputEvent = engine_handle_input;
    engine.app = state;

    /* 音声再生用 engine の生成 */
    createSoundEngine();
    /* 音声プレイヤーオブジェクトの生成 */
    createAssetAudioPlayer(&engine);

    /* このアプリケーションのメインループ */
    while (1) {
        int ident;
        int events;
        struct android_poll_source* source;

	/*
	 * ALooper_pollAll() にてイベントのポーリングを行い、
	 * 溜まっているイベント全てを処理するまでループします。
	 */
        while ((ident=ALooper_pollAll(engine.needToDraw ? 0 : -1, NULL, &events,
                (void**)&source)) >= 0) {

            // Process this event.
            if (source != NULL) {
                source->process(state, source);
            }

            if (state->destroyRequested != 0) {
                __android_log_print (ANDROID_LOG_INFO, LOG_TAG, "Engine thread destroy requested!");
		/*
		 * 終了要求のイベントがあった場合は終了処理の関数を呼び出します。
		 */
                engine_term_display(&engine);
                return;
            }
        }

        if (engine.needToDraw) {
	    /*
	     * 画面の描画が必要と判断された場合は描画関数を呼び出します。
	     */
            engine_draw_frame(&engine);
        }
    }
}
